/*
						 ***********
						** Summary **
						 ***********
	(1)	Title:
				Atlas Browser

	(2)	Purpose:
			To browse the Daggerfall game world maps, displaying the
			'thumbnail' guide embedded in each place.

	(3)	Non-ANSI features:
*/
#ifdef __GNUC__
#	pragma pack(1)
#endif
/*
	(4)	Platform dependencies:
			***  None  ***

	(5)	Change history:
			***  None  ***

	(6)	Input:

			Structures:
				***  None  ***

			Common:
				***  None  ***

			Files:
				arena2/maps.bsa		--	The catalogue of Daggerfall
										locations.

			Assertions about input:

		Every Daggerfall location is catalogued with a sort of thumbnail
		portrait revealing interesting features about the place.  Use
		the atlas tool to browse the location thumbnail of a Daggerfall
		location in any province.

		Usage:	atlas  aProvinceName  aPlaceName
					<shows interesting features of a place>
				atlas  aProvinceName
					<shows all place names in a province>
					<but no thumbnails>
				atlas  aProvinceName  *
					<shows all the thumbnails in a province>
				atlas
					<shows all province names>

		If atlas is installed in a directory different from the
		Daggerfall game directory (typically c:\dagger), define the
		MAP variable in the environment to point to the game
		directory.  E.g.

			set map=g:\game\dag
			atlas anticlere

		The place thumbnail displayed uses the usual compass rose with
		North topmost on the screen.  A location thumbnail vaguely
		resembles a page from a train or bus schedule, showing features
		of interest in as compact a form as possible.

		Names can be abbreviated to the fewest number of letters to
		make them distinct.  So `atlas dag dag' produces the same
		report as `atlas Daggerfall Daggerfall'.

		Place name matching tries first to match the user's place name
		at the start of a name, and if that fails, tries to match it at
		the end of a name.

		So `atlas daggerfall gothway' finds Gothway, Gothway Hall, and
		Gothway Garden.  While `atlas daggerfall Manor' finds all the
		place names ending `...Manor'.

		Using the special form * for the desired place name will display
		the thumbnails for all the locations in the selected province.
		The output from this command can be large.  For example,
		Daggerfall province has over 1300 locations.


	(7)	Output:

			Structures:
				***  None  ***

			Common:
				***  None  ***

			Files:
				***  None  ***

			Assertions about the output:
				Note that the report output may be redirected to a file,
				or piped to another command.

				Note also that the report commentary documents the
				quest scenario command to add that place to the
				overhead map, if it isn't already present.

	(8)	Public members:

			Title:

			Purpose:

			Parameters:

	(9)	Private members:

			Title:

			Purpose:

			Parameters:

	(10) Subprograms called:
			***  None  ***

	(11) Includes referenced:
			***  None  ***

	(12) Structures declared:
			ColorCode				--	Map a site color code to text.
			DItemKey				--	MapDItem directory element.
			FactionCode				--	Map a faction number to text.
			JumpLink				--	Quest spot/teleport cheat element.
			Map						--	Catalogue entry descriptor.
			MapEntry				--	Directory entry for a BSA archive.
			MapItem					--	Map substructure for DItem/PItem.

						 ****************
						** Declarations **
						 ****************

					 ***********************
					** Global declarations **
					 ***********************

				  Copyright 2001, Nagual Enterprises
			Please see COPYING for the complete GNU license

					  ********************
					 ** Global functions **
					  ********************
*/

#			ifndef __BORLANDC__
#				include <unistd.h>
#			else
#				include <io.h>
#				include <dos.h>
#				define __inline__
#			endif
#				include <stdlib.h>
#				include <stdio.h>
#				include <errno.h>
#				include <ctype.h>
#				include <string.h>
#				include <sys/types.h>
#				include <sys/stat.h>
#				include <fcntl.h>
				extern char* stpcpy(char* dest, const char* src);

/*
					  *************
					 ** Constants **
					  *************
*/
					/*	Define some handy mnemonics insulating us from
						os vagaries. */

#			ifdef MSDOS
#				define binaryI				(O_RDONLY | O_BINARY)
#			elif __BORLANDC__
#				define binaryI				(O_RDONLY | O_BINARY)
#			else
#				define binaryI				O_RDONLY
#			endif
#				define true					1
#				define false				0

					/*	Define some convenience constants for
						index mapping. */

#			ifdef __BORLANDC__
#				define logInt				1
#				define WhoAmI				"ATLAS"
#			else
#				define logInt				2
#				define WhoAmI				"atlas"
#			endif
#				define logShort				1
#				define logLong				2

/*
					   **********
					  ** Macros **
					   **********
*/

#				define isNil(z)				(!(z))
#				define notNil(x)			(!isNil(x))
#				define isEquals(x, y)		(!strcmp(x, y))
#				define isEqual(x, y, z)		(!strncmp(x, y, z))
#			ifdef __BORLANDC__
#				define DirSeperator			'\\'
#				define sortString(x, y)		(stricmp(x, y))
#				define isEqualIs(x, y)		(!stricmp(x, y))
#				define isEqualI(x, y, z)	(!strncmpi(x, y, z))
#				define BitBucket			"nul"
#			else
#				define DirSeperator			'/'
#				define sortString(x, y)		(strcasecmp(x, y))
#				define isEqualIs(x, y)		(!strcasecmp(x, y))
#				define isEqualI(x, y, z)	(!strncasecmp(x, y, z))
#				define BitBucket			"/dev/null"
#			endif

				/*	Converters to various metrics of interest. */

					/*	Convert a mapname offset to a mappitem
						table offset. */

#				define nameToPItem(z)		(((z) - 4) / 8)

					/*	Convert a mapname offset to a place
						index number. */

#				define nameToIndex(z)		(((z) - 4) / 32)
#				define indexToName(z)		((z) * 32 + 4)

					/*	Convert a mapname offset to its corresponding
						maptable offset. */

#				define nameToTable(z)		(nameToIndex(z) * 17)
#				define indexToTable(z)		(17 * (z))

					/*	Convert a place type to text. */

#				define placeType(z)			(PlaceTypes[(z)&31])

					/*	Convert a faction number representation. */

#				define factionToName(z)		(asFactionCode(z)->alias)
#				define factionToBlock(z)	(asFactionCode(z)->block)

/*
					 **************
					** Structures **
					 **************
*/
					/*	Define the common structure between dungeon
						and regular sites. */

				struct _MapItem
				{
					long offset;			/*	Offset to this item. */
					long doorList;			/*	Offset to door list for item. */
					short doorCount;		/*	Size of the door list array. */
					short preCount;			/*	Size of the pre-list array. */
					long preamble;
					int itemCount;			/*	Count of current items. */
					int handle;
					unsigned short
						location;			/*	Permanet site index. */
				};
				typedef struct _MapItem MapItem;
#				define newMapItemA(z)	((MapItem*)malloc((z)*sizeof(MapItem)))
#				define newMapItem(z)	((MapItem*)malloc(z))

					/*	Towns and such have a list of doors leading to
						interiors (shops, homes, guilds, etc).  nameHash
						ought to encode the shop name, etc, but how it
						does so is presently unknown. */

				struct _DoorStop
				{
					unsigned short nameHash;
					char name[16];			/*	Always 0 in maps.bsa. */
					short faction;			/*	Faction housed here. */
					long location;			/*	The regular place hash. */
					char place;				/*	Place code. */
					char style;				/*	Shop rating on a 20 point
												scale.  Reduction by 4 seems
												to map onto the `rusty relics'
												and `incense' ratings. */
				};
				typedef struct _DoorStop DoorStop;
#				define newDoorStopA(z)	((DoorStop*)malloc((z)*sizeof(DoorStop)))
#				define newDoorStop(z)	((DoorStop*)malloc(z))

					/*	The mapDItem directory for a province is almost
						sorted by place key, unless the province happens
						to have a "permanent" dungeon.  These all seem to
						have the key 0xc3nn, which corrupts strict serial
						order. */

				struct _DItemKey
				{
					long offset;			/*	Relative displacement to
												data.  Counts from 0 after
												the mapDItem directory table. */
					short one;				/*	Always 1. */
					unsigned short key;		/*	Characteristic location id. */
				};
				typedef struct _DItemKey DItemKey;
#				define newDItemKeyA(z)	((DItemKey*)malloc((z)*sizeof(DItemKey)))
#				define newDItemKey(z)	((DItemKey*)malloc(z))

					/*	Define a container to arbitrate the mapping. */

				struct _Map
				{
					int mapsBsa;			/*	Handle to maps.bsa */
					long mapName;			/*	Offset to mapnames.??? */
					MapItem pItem;			/*	Offset to mappitem.??? */
					MapItem dItem;			/*	Offset to mapditem.??? */
					long mapTable;			/*	Offset to maptable.??? */
					int province;			/*	Currently selected province #. */
					long nameSize;			/*	Size of the current place name
												file. */
					int placeIndex;			/*	Index of the current place. */
					short blockWidth;		/*	Width (in map blocks) of
												current place. */
					short blockHeight;		/*	Height (in map blocks) of
												current place. */
					char fileBlockIndex[64];/*	Current map block grid points. */
					char* provName;			/*	Current province name. */
					char* placeName;		/*	Current place name. */
					char* mapDir;			/*	Path to maps.bsa */
					char* mapNames;			/*	Set of current place names. */
					char* mapSizes;			/*	Set of name lengths. */
					char* dItems;			/*	DItem directory. */
					DoorStop* factions;		/*	Factions represented here. */
					int colorCode;			/*	Site color on province map. */
				};
				typedef struct _Map Map;
#				define newMapA(z)		((Map*)malloc((z)*sizeof(Map)))
#				define newMap(z)		((Map*)malloc(z))

					/*	Declare a BSA member element record.  Alias the
						member name extension to an integer to speed up
						province identification matching. */

				struct _MapEntry
				{
					char name[8];			/*	Member filename part. */
					long prov;				/*	Alias to member extension
												part. */
					short fill;				/*	Empty half-word. */
					long fSize;				/*	Size of this archive member. */
				};
				typedef struct _MapEntry MapEntry;
#				define newMapEntryA(z)	((MapEntry*)malloc((z)*sizeof(MapEntry)))
#				define newMapEntry(z)	((MapEntry*)malloc(z))

					/*	Every place has an optional list of
						links within it to `jump' to or put
						quest items at.  (In dungeons, these
						can be reached using the '[' or ']'
						cheat keys) */

				struct _JumpLink
				{
					short door;				/*	Door list index matching
												this sector tag.  -1 if
												there is no corresponding
												door (e.g. dungeon teleport) */
					short flags;
					short sector;			/*	Sector tag.  In conjunction with
												the usual short location code,
												forms the fully qualified save
												game address of this sector.

												Teleport jump sectors are
												encoded by 0xfann, where nn
												is the position of the sector
												along the jump chain. */
				};
				typedef struct _JumpLink JumpLink;
#				define newJumpLinkA(z)	((JumpLink*)malloc((z)*sizeof(JumpLink)))
#				define newJumpLink(z)	((JumpLink*)malloc(z))

					/*	Define an association for named 16 bit entities. */

				struct _ShortCode
				{
					short code;				/*	16 bit entity. */
					char* alias;			/*	Name for this code. */
				};

					/*	A ColorCode maps a maptable color field to plain
						text. */

				typedef struct _ShortCode ColorCode;
#				define newColorCodeA(z)	((ColorCode*)malloc((z)*sizeof(ColorCode)))
#				define newColorCode(z)	((ColorCode*)malloc(z))

					/*	A FactionCode maps a faction id to plain text. */

				struct _FactionCode
				{
					short code;				/*	16 bit entity. */
					short block;			/*	Thumbnail block type.  -1 if
												there is no unique block type
												(e.g. The Merchants) */
					char* alias;			/*	Name for this code. */
				};
				typedef struct _FactionCode FactionCode;
#				define newFactionCodeA(z)	((FactionCode*)malloc((z)*sizeof(FactionCode)))
#				define newFactionCode(z)	((FactionCode*)malloc(z))

/*
					   **********
					  ** Common **
					   **********
*/

				char* PlaceTypes[]={
					"Alchemist",
					"Town1",
					"Armourer",
					"Bank",
					"Town4",
					"Book store",
					"Clothing shop",
					"Furniture shop",
					"Jeweler",
					"General store",
					"Library",
					"Guild hall",
					"Pawn shop",
					"Weapon smith",
					"Temple",
					"Tavern",
					"Palace",
					"House1",
					"House2",
					"House3",
					"House4",
					"House5",
					"House6",
					"Town23",
					NULL,
				};
#				define DoOvers			16
#				define StartOver		25
#				define StartOverSize	15

					/*	Reverse-index of place types to block file indices.
						Negative values indicate special handling is
						required. */

				int TypeToBlock[]={
					5, -3, 4, 6, -3, 7, 8, 9, 10, 1,
					11, -4, 12, 3, -2, 0, 15, -3, -3, -3,
					-3, -3, -3, -3, -1
				};
				char* Temples[16];
				char* Houses[16];

					/*	The color code has a fine structure as well.
						When a new character is rolled, the sites tagged
						0x8n are rolled against to determine which of them
						are eligible to include on the overhead map from
						the beginning.  All other 0x8n sites are initially
						hidden on the map.  Conversely, 0xan sites are never
						obscured on the overhead map.  0xad (not shown) is the
						code assigned to the ship sites. */

				ColorCode siteColors[]={
					0x84, "orange",
					0x87, "dark orange",
					0x8a, "medium red",
					0x8c, "dark red",
					0x8d, "black",
					0xa0, "light rose",
					0xa1, "medium rose",
					0xa2, "medium rose",
					0xa3, "medium rose",
					0xa5, "light blue",
					0xa6, "dark rose",
					0xa8, "light pink",
					0xa9, "dark blue",
					0xab, "dark pink",
					0xac, "dark pink",
					0, NULL
				};

				char* MapDir = "";

					/*	Enumerate the available province names. */

				char* MapProvinces[]={
					"Alik'r Desert",
					"Dragontail Mountains",
					"",
					"",
					"",
					"Dwynnen",
					"",
					"",
					"",
					"Isle of Balfiera",
					"",
					"Dak'fron",
					"",
					"",
					"",
					"",
					"Wrothgarian Mountains",
					"Daggerfall",
					"Glenpoint",
					"Betony",
					"Sentinel",
					"Anticlere",
					"Lainlyn",
					"Wayrest",
					"",
					"",
					"Orsinium Area",
					"",
					"",
					"",
					"",
					"",					/*	31 houses the Mantellan Crux
											dungeon, as well as the small
											and large ships. */
					"Northmoor",
					"Menevia",
					"Alcaire",
					"Koegria",
					"Bhoriane",
					"Kambria",
			 		"Phrygias",
					"Urvaius",
					"Ykalon",
					"Daenia",
					"Shalgora",
					"Abibon-Gora",
					"Kairou",
					"Pothago",
					"Myrkwasa",
					"Ayasofya",
					"Tigonus",
					"Kozanset",
					"Satakalaam",
					"Totambu",
					"Mournoth",
					"Ephesus",
					"Santaki",
					"Antiphyllos",
					"Bergama",
					"Gavaudon",
					"Tulune",
					"Glenumbra Moors",
					"Ilessan Hills",
					"Cybiades",
					NULL
				};

					/*	Enumerate the map block grid points of interest. */

				char* BlockNames[]={
					"Tavern",
					"General store",
					"Residence",
					"Weapon smith",
					"Armourer",
					"Alchemist",
					"Bank",
					"Book store",
					"Clothing store",
					"Furniture store",
					"Jeweler",
					"Library",
					"Pawn shop",
					"Temple",
					"Temple",
					"Palace",
					"Farm",
					"Dungeon",
					"Castle",
					"Manor",
					"Shrine",
					"Ruin",
					"Shack",
					"Cemetery",
					"Empty lot",
					"Order of the Raven",
					"Knights of the Dragon",
					"Knights of the Owl",
					"Knights of the Moon",
					"Order of the Candle",
					"Knights of the Flame",
					"Host of the Horn",
					"Knights of the Rose",
					"Knights of the Wheel",
					"Order of the Scarab",
					"Knights of the Hawk",
					"Mage guild",
					"Thief guild",
					"Dark Brotherhood",
					"Fighter guild",
					"Custom area",
					"City wall",
					"Shopping mall",
					"Ship",
					"Witch coven",
					"Town1",
					"Town4",
					"Town23",
					NULL
				};
				int BlockCounts[49];

				/*	Enumerate the factions appearing in maps.bsa */

#				define EmptyFaction		29
				FactionCode MapFactions[]={
					{ 21, -1, "Arkay"},
					{ 22, -1, "Zen"},
					{ 24, -1, "Mara"},
					{ 27, -1, "Julianos"},
					{ 29, -1, "Dibella"},
					{ 33, -1, "Stendarr"},
					{ 35, -1, "Kynareth"},
					{ 40, 36, "Mage's Guild"},
					{ 41, 39, "Fighter's Guild"},
					{ 42, 37, "Thief's Guild"},
					{ 60, -1, "The Academics"},
					{ 84, -1, "The Resolution of Z'en"},
					{ 92, -1, "The Akatosh Chantry"},
					{108, 38, "The Dark Brotherhood"},
					{201, -1, "Daggerfall"},
					{204, -1, "Sentinel"},
					{207, -1, "Wayrest"},
					{368, 26, "Knights of the Dragon"},
					{408, 29, "Order of the Candle"},
					{409, 32, "Knights of the Rose"},
					{410, 30, "Knights of the Flame"},
					{411, 31, "Host of the Horn"},
					{413, 27, "Knights of the Owl"},
					{414, 25, "Order of the Raven"},
					{415, 33, "Knights of the Wheel"},
					{416, 34, "Order of the Scarab"},
					{417, 35, "Knights of the Hawk"},
					{510, -1, "The Merchants"},
					{852, -1, "Random Ruler"},
					{0x7fff, -1, NULL}
				};

/*
				 **********************
				** Local declarations **
				 **********************

					  *************
					 ** Variables **
					  *************

					***  None  ***

						 ********
						** Text **
						 ********


							 *******************
							** Private methods **
							 *******************
*/
		/*	Borland C lseek() is b0rken for file sizes in the
			range 32k-64k, which is where the mapnames sizes
			happen to fall.  Redefine lseek to work right.
			Note that the compiler will crab the method should
			return a value, and it does answer the long result
			in AX:DX, but the compiler has no extension to
			indicate this requirement. (Using the obvious
			`return _AX' apparently cues it that it may freely
			trash DX, and it happily does so)  Note also
			that the coding is tuned for the compact memory
			model. */

#ifdef __BORLANDC__
#	define _asm asm
long Seek
(	int handle,
	long off,
	int how)
{
	_asm {
		lea di,off
		mov dx,ss:[di]
		mov cx,ss:[di+2]
		mov ax,how
		mov ah,42h
		mov bx,handle
		int 21h
	}
	return;
}
#else
#	define Seek(x,y,z)		lseek(x, y, z);
#endif

/*
							 ******************
							** Public methods **
							 ******************
*/

FactionCode* asFactionCode
(	short faction)
{
		/*	Answer the FactionCode with the id of faction,
			or nil if no matching faction is found. */

	FactionCode* facts = MapFactions;

	while (faction >= facts->code)
	{
		if (faction == facts->code)
			return facts;
		facts++;
	}
	return &MapFactions[EmptyFaction];
}

int provToIndex
(	char* prov)
{
		/*	Answer the province number for a province name.  Answer -1
			if the requested name cannot be matched. */

	int provNo = 0;
	char** provs = MapProvinces;
	char first = toupper(*prov);

	/*	If the entry already looks numeric, just convert it to
		binary. */

	if (isdigit(first))
		return atoi(prov);
	while (*provs)
	{
		if (**provs == first)
		{
			/*	Use loose name matching, so the first entry matching
				as many characters as the user provided wins the
				race.  There may be multiple matchs, but we ignore
				the rest.  Strictly speaking, we should count
				*provs too and compare only the mininum, but the
				risk of false positives is slim to none. */

			if (isEqualI(prov, *provs, strlen(prov)))
				return provNo;
		}
		provNo++;
		provs++;
	}
	return -1;
}

Map* openProvince
(	char* prov)
{
		/*	Open the map files associated with a province.  Answer the
			initialized Map structure on success and nil otherwise. */

	long seed = 4L;
	char name[1024];
	long dirLen;
	int provNo = provToIndex(prov);
	Map* theMap;
	short count;
	short found = 0;
	char provId[]=".000";
	long* lProv = (long*)provId;
	div_t province;

	if (provNo < 0)
	{
		printf("%s not found.\n", prov);
		return NULL;
	}
	theMap = (Map*)malloc(sizeof(Map));
	memset(theMap, 0, sizeof(Map));
	theMap->provName = prov;
	theMap->mapDir = MapDir;
	theMap->province = provNo;

	/*	Build the full path to the maps. */

	dirLen = strlen(theMap->mapDir);
	strncpy(name, theMap->mapDir, dirLen);
	strcpy(&name[dirLen], "maps.bsa");
	if ((theMap->mapsBsa = open(name, binaryI)) < 0)
	{
		perror("atlas");
		printf("Unable to open %s for reading.\n", name);
		free(theMap);
		return NULL;
	}
	theMap->pItem.handle = theMap->dItem.handle = theMap->mapsBsa;

	/*	Scan the archive directory for the map files associated with
		the province of interest.  The member directory is appended
		at the end of the maps.bsa file, so scroll to that point to
		begin searching for map names. */

	province = div((int)provNo, 10);
	provId[2] = province.quot + 48;
	provId[3] = province.rem + 48;
	read(theMap->mapsBsa, &count, sizeof(count));
	dirLen = Seek(theMap->mapsBsa, 0L, SEEK_END);
	dirLen -= (long)(18*count);
	dirLen = Seek(theMap->mapsBsa, dirLen, SEEK_SET);

	/*	Inject the seed offset until all the province files are located. */

	while (count--)
	{
		MapEntry me;
		read(theMap->mapsBsa, me.name, 18);
		if (*lProv == me.prov)
		{
			/*	This file matchs the province we're interested in.
				Record the offset to this data area for reuse. */

			if (me.name[3] == 'N')
			{
				/*	mapnames.??? */

				theMap->nameSize = me.fSize;
				theMap->pItem.itemCount = nameToIndex(me.fSize);
				theMap->mapName = seed;
			}
			else if (me.name[3] == 'P')
				/*	mappitem.??? */

				theMap->pItem.offset = seed;
			else if (me.name[3] == 'T')
				/*	maptable.??? */

				theMap->mapTable = seed;
			else if (me.name[3] == 'D')
				/*	mapditem.??? */

				theMap->dItem.offset = seed;
			if (++found > 3)
				return theMap;
		}
		seed = seed + me.fSize;
	}
	printf("Unable to locate province in maps.bsa.\n");
	return NULL;
}

long nameOffset
(	Map* theMap,
	long startingOffset)
{
		/*	Answer the mapname offset of the requested place.
			Answer -1 if the requested place cannot be located.
			Start scanning mapname entries from the requested
			offset. */

	int count;
	int pLen;
	char* names;
	char* sizes;
	long size = 0L;
	char first;
	char isWild = false;

	size = theMap->nameSize;
	count = theMap->pItem.itemCount;
	if (!theMap->mapNames)
	{
		/*	On first access, acquire the needed storage for
			place names and populate it. */

		long dCount;
		names = (char*)malloc(size);
		sizes = (char*)malloc(count);
		Seek(theMap->mapsBsa, theMap->mapName, SEEK_SET);
		read(theMap->mapsBsa, names, size);
		theMap->mapNames = names;
		theMap->mapSizes = sizes;
		memset(sizes, 0, count);
		Seek(theMap->mapsBsa, theMap->dItem.offset, SEEK_SET);
		read(theMap->mapsBsa, &dCount, sizeof(dCount));
		theMap->dItem.itemCount = dCount;
		theMap->dItems =(char*)malloc((int)dCount << 3);
		read(theMap->mapsBsa, theMap->dItems, (int)dCount << 3);
		printf("-- %d sites, %d dungeons in province %s (%d).\n",
				count,
				(int)dCount,
				MapProvinces[theMap->province],
				theMap->province);
	}
	else
	{
		names = theMap->mapNames;
		sizes = theMap->mapSizes;
	}
	/*	If we need only to initialize the map structure, the current
		place name will be empty, so just return in that case. */

	if (!theMap->placeName)
		return -1L;
	isWild = *theMap->placeName == '*';

	/*	Scan the table of names, looking for matchs.  Populate
		the table of name lengths to speed up subsequent use. */

	pLen = strlen(theMap->placeName);
	pLen = pLen > 31 ? 31 : pLen;
	first = toupper(*theMap->placeName);
	names = &names[startingOffset];
	count -= nameToIndex(startingOffset);
	while (count-- > 0)
	{
		if (isWild)
			return names - theMap->mapNames;
		if (!*sizes)
			*sizes = strlen(names);
		if (*names == first)
		{
			/*	Use loose matching so the user need give only
				enough of the name to identify it uniquely.  Echo
				the full name so he can confirm the match. */

			if (isEqualI(names,
						theMap->placeName,
						(pLen <= *sizes ? pLen : (int)*sizes)))
				return names - theMap->mapNames;
		}
		/*	Since anchoring the match at the head didn't work, try an
			anchor at the tail (e.g., *Manor). */

		if (pLen <= *sizes)
		{
			if (isEqualI(&names[*sizes - pLen], theMap->placeName, pLen))
				return names - theMap->mapNames;
		}
		sizes++;
		names = &names[32];
	}
	return -1L;
}

long dItemOffset
(	Map* theMap,
	long offset)
{
		/*	Answer the dItem offset from the supplied mapPItem offset. */

	int count;
	DItemKey* dir = (DItemKey*)theMap->dItems;

	count = theMap->dItem.itemCount;

	/*	Scan the table of dungeon sites, looking for a matching key. */

	while (count--)
	{
		if (theMap->pItem.location == dir->key)
			return 4 + dir->offset + (theMap->dItem.itemCount << 3);
		dir++;
	}
	return -1L;
}

__inline__ long pItemOffset
(	Map* theMap,
	long offset)
{
		/*	Answer the pItem offset from the supplied mapnames offset. */

	long pItem;

	Seek(theMap->mapsBsa, theMap->pItem.offset + nameToPItem(offset), SEEK_SET);
	read(theMap->mapsBsa, &pItem, sizeof(pItem));
	return pItem + (theMap->pItem.itemCount << 2);
}

__inline__ long preambleSize
(	MapItem* theMap,
	long offset)
{
		/*	Answer the size of the preamble records for a pItem at an offset. */

	long count;
	theMap->preamble = theMap->offset + offset;
	Seek(theMap->handle, theMap->preamble, SEEK_SET);
	read(theMap->handle, &count, sizeof(count));
	theMap->preCount = count;
	return count * 6;
}

__inline__ int doorSize
(	MapItem* theMap,
	long offset)
{
		/*	Answer the size of the door records for a pItem at an offset.
			Door records document the sector addresses (the entities stored
			in savetree.dat) of all the enterable 3D objects on this map.
			They also encode the type of 3D object (guild hall, temple, etc)
			and, possibly, the place name of the object. */

	long skip = preambleSize(theMap, offset);
	short count;

	/*	Look up the key code for this pItem. */

	Seek(theMap->handle, theMap->preamble + skip + 37, SEEK_SET);
	read(theMap->handle, &theMap->location, sizeof(theMap->location));

	skip += 71 + 32 + 9 + 4 + offset;
	Seek(theMap->handle, theMap->offset + skip, SEEK_SET);
	read(theMap->handle, &count, sizeof(count));
	theMap->doorList = skip;
	theMap->doorCount = count;
	return (int)count;
}

__inline__ int fileBlocks
(	Map* theMap,
	long pItem)
{
		/*	Answer the file block index values for the pItem.
			File blocks index the member directory of blocks.bsa
			and represent an x by y grid detail of the place
			of interest. */

	long doors = doorSize(&theMap->pItem, pItem);
	long reveal = 0;
	int color;
	ColorCode* cCode = siteColors;

	doors *= 26;
	doors += 5 + 32 + 8 + 2 + theMap->pItem.doorList;
	Seek(theMap->mapsBsa, theMap->pItem.offset + doors, SEEK_SET);
	read(theMap->mapsBsa, &theMap->blockWidth, 1);
	read(theMap->mapsBsa, &theMap->blockHeight, 1);
	doors += theMap->pItem.offset + 2 + 7;
	printf(" at %#lx.\n", doors);
	Seek(theMap->mapsBsa, doors, SEEK_SET);
	read(theMap->mapsBsa,
		 theMap->fileBlockIndex,
		 sizeof(theMap->fileBlockIndex));

	/*	And, the point of this whole exercise, fabricate a quest
		scenario template to reveal this place on the province
		map. */

	Seek(theMap->mapsBsa,
		 theMap->mapTable + indexToTable(theMap->placeIndex),
		 SEEK_SET);
	read(theMap->mapsBsa, &reveal, 4);
	reveal &= 0xfffff;
	printf(	"-- Quest source:  reveal _aPlace_ in province %d at %#lx\n",
			theMap->province,
			reveal);

	/*	Look up the map color for this item. */

	Seek(theMap->mapsBsa, 3L, SEEK_CUR);
	read(theMap->mapsBsa, &color, sizeof(color));
	theMap->colorCode = (color & 0xffff) >> 1;
	printf("-- Site: 0x%04x, color: ", theMap->pItem.location);
	while (cCode->code)
	{
		if (theMap->colorCode < cCode->code)
			break;
		if (cCode->code == theMap->colorCode)
		{
			printf("%s", cCode->alias);
			if ((theMap->colorCode&0x20) != 0x20)
				printf(" (may be absent initially)");
			printf("\n");
			return 1;
		}
		cCode++;
	}
	/*	If no color code translation was found, just emit the raw
		hex code. */

	printf(" %#x\n", theMap->colorCode);
	return 1;
}

int listProvs
()
{
		/*	List the available provinces. */

	char** provs = MapProvinces;
	while (*provs)
	{
		if (**provs)
			printf("%s\n", *provs);
		provs++;
	}
	return 0;
}

int sortMapName
(	const void* e1,
	const void* e2)
{
		/*	qsort() appendage to sort place names ignoring case. */

	return sortString((char*)e1, (char*)e2);
}

int listNames
(	Map* theMap)
{
		/*	Produce a sorted listing of place names on the map.
			Note that mapname sizes all fit within 64k so no
			memory model issues arise in Dos-land. */

	int count;
	char* names, *nPtr;
	count = theMap->pItem.itemCount;
	names = (char*)malloc(theMap->nameSize);
	memcpy(names, theMap->mapNames, theMap->nameSize);
	nPtr = &names[4];
	qsort(nPtr, count, 32, sortMapName);
	while (count--)
	{
		printf("%s\n", nPtr);
		nPtr = &nPtr[32];
	}
	free(names);
	return 0;
}

int preambleLinks
(	MapItem* item)
{
		/*	Count the links (teleport jumps) in this preamble. */

	int links = 0;
	int count = item->preCount;
	JumpLink jl;

	Seek(item->handle, item->preamble + 4, SEEK_SET);
	while (count--)
	{
		read(item->handle, &jl, sizeof(jl));
		if (((int)jl.sector & 0xff00) == 0xfa00)
			links++;
	}
	return links;
}

int printItem
(	MapItem* item)
{
		/*	Print a MapItem header. */

	int links;

	printf(	"--   %d preambles at %#lx\n",
			item->preCount,
			4 + item->preamble);
	printf(	"--   %d doors at %#lx\n",
			item->doorCount,
			7 + item->doorList + item->offset);
	return 1;
}

int printRating
(	DoorStop* door)
{
		/*	Print the shop rating using a 4 star system:
				`incense'			--	****
				`better appointed'	--	***
				`practical'			--	**
				'cobbled'			--	*
				'rusty'				--
			The point scale is actually 21 (from 0 to 20 inclusive). */

	int stars = (door->style - 1) >> 2;
	int rem;

	stars = stars < 0 ? 0 : stars;
	rem = 4 - stars;
	while (rem--)
		printf(" ");
	while (stars--)
		printf("*");
	return 1;
}

int printDoor
(	DoorStop* door,
	int index)
{
	char* factName;

	if (door->faction > 0)
	{
		printf(	"[%3d]   %04hx %5ld ",
				index,
				door->nameHash,
				door->location&0xffff);
		printRating(door);
		printf("%s", placeType(door->place));
		factName = factionToName(door->faction);
		if (door->faction != 510 && factName)
		{
			printf(	", %s\n", factName);
		}
		else
		{
			if (door->faction != 510)
				printf(" (%hd)\n", door->faction);
			else if (door->faction == 510 &&
				(door->place == 1 ||
				door->place == 4 ||
				door->place > 16))
				printf(	", The Merchants\n");
			else
				printf("\n");
		}
	}
	else
		printf("\n");
	return 1;
}

int countDoors
(	Map* theMap)
{
		/*	Count exactly the different types of shops. */

	int count = theMap->pItem.doorCount;
	DoorStop* door = theMap->factions;
	char** temples = Temples;
	char** houses = Houses;
	int* blocks = &BlockCounts[13];

	if (!count)
		return 0;
	memset(BlockCounts, 0, DoOvers*sizeof(int));
	memset(&BlockCounts[StartOver], 0, StartOverSize*sizeof(int));
	memset(Temples, 0, sizeof(Temples));
	memset(Houses, 0, sizeof(Houses));
	while (count--)
	{
		int block = TypeToBlock[door->place];
		switch (block)
		{
		case -4:
			BlockCounts[factionToBlock(door->faction)]++;
			break;
		case -3:
			if (door->faction)
			{
				FactionCode* faction = asFactionCode(door->faction);
				char* house = faction->alias;
				if (house != MapFactions[9].alias &&
					house != MapFactions[13].alias)
				{
					*houses++ = house;
					break;
				}
				BlockCounts[faction->block]++;
			}
			else
				BlockCounts[2]++;
			break;
		case -2:
			*temples++ = factionToName(door->faction);
			if (!*blocks)
				(*blocks)++;
			blocks++;
			break;
		case -1:
			break;
		case 15:
			if (BlockCounts[block])
				break;
		default:
			BlockCounts[block]++;
			break;
		}
		door++;
	}
	return 1;
}

int printTranspose
(	Map* theMap)
{
		/*	Print the grid point highlights for the currently selected
			place.

			Note that the grid is stored southwest to northeast, and we
			(un)ravel that to the conventional compass rose setting with
			north topmost on the display screen. */

	int rows = theMap->blockHeight;
	char* fbi = &theMap->fileBlockIndex[(rows-1)*theMap->blockWidth];
	int* bc = BlockCounts;
	int id = 0;
	int total = 0;
	char** blockType = BlockNames;
	char** houses = Houses;

	memset(BlockCounts, 0, sizeof(BlockCounts));
	printf("\n ");
	while (rows--)
	{
		char* old = fbi;
		int col = theMap->blockWidth;
		while (col--)
		{
			printf(" %02d", (int)*fbi);
			BlockCounts[*fbi++]++;
		}
		fbi = &old[-theMap->blockWidth];
		printf("\n ");
	}
	/*	Finally display the legend, plus use-counts, for the elements
		on the current map grid. */

	printf("\n");
	countDoors(theMap);
	while (*blockType)
	{
		if (*bc)
		{
			printf("%3d %s", *bc, *blockType);
			if (*bc > 1 && (id < 24 || id > 35))
				printf("s");
			switch (id)
			{
			case 13:
				printf(", %s (%d)\n", Temples[0], id);
				break;
			case 14:
				printf(", %s (%d)\n", Temples[1], id);
				break;
			case 45:
			case 46:
			case 47:
				printf("\n");
				break;
			default:
				printf(" (%d)\n", id);
				break;
			}
		}
		total += *bc++;
		id++;
		blockType++;
	}
	/*	Exclude non-door elements from the door count and correct
		for counting the thief and assassin guilds twice. */

	total -= BlockCounts[23] + BlockCounts[40] +
			 BlockCounts[41] + BlockCounts[42];
	while (*houses)
	{
		printf("  1 House, %s\n", *houses++);
		total++;
	}
	if (total)
		printf("\n%3d doors in total.\n", total);
	return 1;
}

DoorStop* getDoors
(	Map* theMap)
{
		/*	Retrieve the doorways present in this area. */

	MapItem* item = &theMap->pItem;
	int count = item->doorCount;
	int index = 0;
	DoorStop* doors;
	int size = sizeof(DoorStop) * item->doorCount;

	if (!count)
		return NULL;
	doors = theMap->factions = newDoorStop(size);
	memset(theMap->factions, 0, size);
	Seek(item->handle, 7 + item->doorList + item->offset, SEEK_SET);
	read(item->handle, doors, size);
	return doors;
}

int printDoors
(	Map* theMap)
{
		/*	Print the summary report of doors in this area. */

	MapItem* item = &theMap->pItem;
	int count = item->doorCount;
	int index = 0;
	int first = 1;
	DoorStop* door = theMap->factions;

	if (!count)
		return 0;
	while (count--)
	{
		if (door->faction > 0)
		{
			if (first)
			{
				printf("   Town Doors (SW to NE)\n\nDoor#  name?  offset  style\n\n");
				first = 0;
			}
			printDoor(door, index);
		}
		door++;
		index++;
	}
	printf("\n");
	return 1;
}

int getMapDir
()
{
		/*	Get the link to the Daggerfall installation directory. */

	char* dir;
	char* path;

	dir = getenv("MAP");
	if (dir)
	{
		int size = strlen(dir);
		char* sPtr;
		path = (char*)malloc(9+size);
		sPtr = stpcpy(path, dir);
		if (sPtr[-1] != DirSeperator)
			*sPtr++ = DirSeperator;
		strncpy(sPtr, "arena2", 6);
		sPtr = &sPtr[6];
		*sPtr++ = DirSeperator;
		*sPtr++ = '\0';
		MapDir = path;
	}
	return dir != NULL;
}

int main
(	int argc,
	char** argv)
{
		/*	Query maps.bsa data sets for interesting features. */

	Map* theMap;
	long nItem;
	int isAtlas = strstr(argv[0], WhoAmI) != NULL;

	getMapDir();

	/*	No arguments?  Just print out a list of province names. */

	if (argc < 2)
		return listProvs();
	theMap = openProvince(argv[1]);
	if (!theMap)
	{
		printf("Exiting, as there is no mapping.\n");
		return 1;
	}
	/*	Just one argument?  Print out all the place names in the
		specified province, but give no thumbnails. */

	if (argc == 2)
	{
		nameOffset(theMap, 4);
		return listNames(theMap);
	}
	/*	Otherwise, print out all the thumbnails matching the given
		place name. */

	theMap->placeName = argv[2];
	nItem = nameOffset(theMap, 4);
	if (nItem < 0)
	{
		printf("Exiting, as there is no place name offset.\n");
		return 1;
	}
	while (nItem >= 0)
	{
		long pItem = pItemOffset(theMap, nItem);

		theMap->placeIndex = nameToIndex(nItem);
		printf(	"-- Found %s (%d)",
				&theMap->mapNames[nItem],
				theMap->placeIndex);
		if (fileBlocks(theMap, pItem))
		{
			long dItem = dItemOffset(theMap, pItem);
			int links = 0;
			int spots = 0;

			if (!isAtlas)
			{
				printf("-- mapPItem at %#x:\n", theMap->pItem.offset);
				printItem(&theMap->pItem);
			}
			if (dItem > 0 && theMap->dItem.itemCount)
			{
				doorSize(&theMap->dItem, dItem);
				if (!isAtlas)
				{
					printf("-- mapDItem at %#x:\n", theMap->dItem.offset);
					printItem(&theMap->dItem);
				}
				links = preambleLinks(&theMap->dItem);
			}
			spots = preambleLinks(&theMap->pItem);
			theMap->factions = getDoors(theMap);
			printTranspose(theMap);
			if (spots)
				printf("%3d quest locations.\n", spots);
			if (links)
				printf("%3d dungeon links.\n", links);
			printf("\n");
			if (!isAtlas)
				printDoors(theMap);
		}
		nItem = nameOffset(theMap, 32 + nItem);
	}
	return 0;
}

